import time
import os
import sys
import pickle


import pandas as pnd
from Bio.KEGG import REST



def download_keggorg(logger, keggorg='lpl', outdir='./', ):


    # check if already downloaded
    outfile = os.path.join(outdir, f'{keggorg}.keggorg')
    if os.path.exists(outfile):
        logger.info(f"Organism code '{keggorg}' already downloaded ('{os.path.join(outdir, f'{keggorg}.keggorg')}').")
        return 0
    
    
    # donwload entire txt:
    logger.info(f"Verifying existence of organism code '{keggorg}' on KEGG...")
    time.sleep(0.5)   # be respectful
    try: response = REST.kegg_list(keggorg).read()
    except: 
        logger.error(f"Organism code '{keggorg}' not found in KEGG database.")
        return 1
    # response is now a string similar to:
    """
    lpl:lp_0026	CDS	31317..32084	hydrolase, HAD superfamily, Cof family
    lpl:lp_0027	CDS	complement(32236..32907)	pgmB1; beta-phosphoglucomutase
    """

    
    # extract the gene IDs list:
    gene_ids = [line.split('\t')[0] for line in response.strip().split('\n')]
    # example of gene_id: "lpl:lp_0005"
    logger.info(f"Respectfully downloading {len(gene_ids)} genes from KEGG...")
    
    
    
    # respectfully download in batch
    # 10 is the max number of elements that can be downloaded
    batch_size = 10
    n_batches = len(gene_ids) // batch_size + (1 if (len(gene_ids) % batch_size) > 0 else 0) 


    n_attempts = 5
    attempts_left = n_attempts
    default_sleep = 0.5
    sleep_time = default_sleep


    completed_batches = 0
    completed_genes = 0
    res_string_list = []
    while completed_batches < n_batches:

        # be respectful
        time.sleep(sleep_time)

        # extract batch 
        start_index = completed_batches *batch_size
        end_index = (completed_batches+1) *batch_size
        if end_index > len(gene_ids): end_index = len(gene_ids)
        curr_batch = gene_ids[start_index: end_index]


        # download batch
        try:
            res_string = REST.kegg_get(curr_batch).read()
            for item in res_string.split("///\n\n"):
                res_string_list.append(item.replace('///\n', ''))
            completed_batches += 1
            completed_genes += len(curr_batch)

            print(f"{completed_genes}/{len(gene_ids)} ({int(completed_genes/len(gene_ids)*100)}%) completed!", end='\r', file=sys.stderr)

            attempts_left = n_attempts
            sleep_time = default_sleep
        except: 
            attempts_left -= 1
            sleep_time = default_sleep *4  # increase sleep time to be more respectful
            logger.warning(f"An error occurred during kegg_get() of batch {curr_batch}. Remaining attempts: {attempts_left}.")

            
            if attempts_left == 0:
                logger.error("No attemps left! Shutting down...")
                return 1


    # hide last progress trace ('sheets_dicts' unused if not in multi-strain mode):
    last_trace = f"{completed_genes}/{len(gene_ids)} ({int(completed_genes/len(gene_ids)*100)}%) completed!"
    whitewash = ''.join([' ' for i in range(len(last_trace))])
    print(whitewash, end='\r', file=sys.stderr)   
    
    
    
    # extract info into a formatted df:
    df = []  # list of dicts, future df
    for entry in res_string_list:

        entry_dict = {}
        curr_header = None

        for line in entry.split('\n'):
            if line == '': continue

            header = line[:12]
            content = line[12:]
            if header != ' '*12:
                curr_header = header

            if curr_header == 'ENTRY       ':
                gid = content.split(' ', 1)[0]
                entry_dict['gid'] = gid

            if curr_header == 'POSITION    ':
                entry_dict['pos'] = content.strip()

            if curr_header == 'ORTHOLOGY   ':
                ko = content.split(' ', 1)[0]
                entry_dict['ko'] = ko

            if curr_header == 'MOTIF       ':
                db, value = content.strip().split(': ', 1)
                entry_dict[db] = value.split(' ')

            if curr_header == 'DBLINKS     ':
                db, value = content.strip().split(': ', 1)
                entry_dict[db] = value.split(' ')

        df.append(entry_dict)
    df = pnd.DataFrame.from_records(df)
    
    
    # save dataframe in the output dir:
    with open(outfile, 'wb') as wb_handler:
        pickle.dump(df, wb_handler)
    logger.info(f"'{outfile}' created!")
        
        
        
    return 0


from pathlib import Path
import pickle
import os


import pandas as pnd



def parse_eggnog(model, eggnog, idcollection_dict):
    
    
    eggnog = pnd.read_csv(eggnog, sep='\t', comment='#', header=None)
    eggnog.columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')
    eggnog = eggnog.set_index('query', drop=True, verify_integrity=True)
    
    
    # PART 1. get KO codes available
    kos_org = set()
    for gid, kos in eggnog['KEGG_ko'].items():
        if kos == '-': 
            continue
        kos = kos.split(',')
        kos = [i.replace('ko:', '') for i in kos]
        for ko in kos: 
            kos_org.add(ko)
            
            
    # PART 2. get reactions in the organism (even the GPR is not complete)
    krs_org = set()
    for kr, kos in idcollection_dict['kr_to_kos'].items(): 
        if any([ko in kos_org for ko in kos]):
            krs_org.add(kr)
    
    
    return krs_org
    
    

def parse_keggorg(keggorg, outdir, idcollection_dict):

    df_keggorg = pickle.load(open(os.path.join(outdir, f'{keggorg}.keggorg'), 'rb'))
    df_keggorg = df_keggorg.set_index('gid', drop=True, verify_integrity=True)

    
    # PART 1. get KO codes available
    kos_org = set([i for i in df_keggorg['ko'] if pnd.isna(i)==False])
    
    
    # PART 2. get reactions in the organism (even the GPR is not complete)
    krs_org = set()
    for kr, kos in idcollection_dict['kr_to_kos'].items(): 
        if any([ko in kos_org for ko in kos]):
            krs_org.add(kr)
    
    
    return krs_org



def parse_taxon(taxon, idcollection_dict):
    
    
    # formatting of --taxon was already verified at startup.
    # also the presence of 'ko_to_taxa' in idcollection_dict was veryfied at startup.
    level, name = taxon.split(':')
           
        
    # PART 1. get KO codes available
    kos_org = set()
    for ko in idcollection_dict['ko_to_taxa'].keys():
        if name in idcollection_dict['ko_to_taxa'][ko][level]:
            kos_org.add(ko)
            
    
    # PART 2. get reactions in the organism (even the GPR is not complete)
    krs_org = set()
    for kr, kos in idcollection_dict['kr_to_kos'].items(): 
        if any([ko in kos_org for ko in kos]):
            krs_org.add(kr)
    
    
    return krs_org

    
    
def check_completeness(logger, model, progress, module, focus, taxon, eggnog, keggorg, idcollection_dict, summary_dict, outdir): 
    # check KEGG annotations in the universe model to get '%' of completeness per pathway/module.
    
            
    # get the reference set of kr codes (all kegg or organism specific): 
    kr_uni = set()
    if keggorg != '-':  # keggorg has precedence
        kr_uni = parse_keggorg(keggorg, outdir, idcollection_dict)
        kr_uni_label = f"organism code '{keggorg}'"
    elif taxon != '-':
        kr_uni = parse_taxon(taxon, idcollection_dict)
        kr_uni_label = f"taxon '{taxon}'"
    elif eggnog != '-':
        for eggfile in eggnog:
            eggset = parse_eggnog(model, eggfile, idcollection_dict)
            kr_uni = kr_uni.union(eggset)
        kr_uni_label = f"{len(eggnog)} eggnog annotations"
    else: 
        kr_uni = idcollection_dict['kr']
        kr_uni_label = "whole KEGG"
    
    
    # get all the 'kr' annotations in the model
    kr_ids_modeled = set()
    for r in model.reactions: 
        if 'kegg.reaction' in r.annotation.keys():
            for kr_id in r.annotation['kegg.reaction']:
                kr_ids_modeled.add(kr_id)
    kr_uni_missing = kr_uni - kr_ids_modeled
    kr_uni_coverage = len(kr_ids_modeled.intersection(kr_uni)) / len(kr_uni) * 100
    logger.info(f"Coverage for {kr_uni_label}: {round(kr_uni_coverage, 0)}% ({len(kr_uni_missing)} missing).")
    
    
    # define the map?????, containing krs not included in maps
    krs_in_maps = set()
    for i in summary_dict: krs_in_maps = krs_in_maps.union(i['kr_ids']) 
    krs_not_in_maps = idcollection_dict['kr'] - krs_in_maps
    summary_dict.append({
        'map_id': 'map?????',
        'map_name': 'Not included in maps',
        'kr_ids': krs_not_in_maps,
        'cnt_r': len(krs_not_in_maps),
        'mds': []
    })    
    
    
    # get all the map / md codes:
    map_ids = set()
    md_ids = set()
    for i in summary_dict:
        map_ids.add(i['map_id'])
        for j in i['mds']:
            md_ids.add(j['md_id'])
            
            
    # check if 'focus' exist
    if focus != '-' and focus not in map_ids and focus not in md_ids:
        logger.error(f"The ID provided with --focus does not exist: {focus}.")
        return 1
    if focus.startswith('map'):
        logger.debug(f"With --focus {focus}, --module will switch to False.")
        module = False
    if focus != '-':
        missing_logger = ()
    
    
    # define some counters:
    maps_finished = set()
    maps_noreac = set()
    maps_missing = set()
    maps_partial = set()

    
    list_coverage  = []
    
    
    # iterate over each map:
    for i in summary_dict:
        
        
        # get ID and name: 
        map_id = i['map_id']
        map_name_short = f"{i['map_name'][:20]}"
        if len(i['map_name']) > 20: 
            map_name_short = map_name_short + '...'
        else:  # add spaces as needed: 
            map_name_short = map_name_short + ''.join([' ' for i in range(23-len(map_name_short))])
            
            
        # check if this map was (at least partially) covered:
        map_krs = set([kr for kr in i['kr_ids'] if kr in kr_uni])
        missing = map_krs - kr_ids_modeled
        present = kr_ids_modeled.intersection(map_krs)
        if focus == map_id: 
            missing_logger = (map_id, missing)

        
        # put the map in the right bucket:
        if missing == set() and map_krs != set():
            maps_finished.add(map_id)
        elif map_krs == set():
            maps_noreac.add(map_id)
        elif missing == map_krs:
            maps_missing.add(map_id)
        elif len(missing) < len(map_krs):
            maps_partial.add(map_id)
            
            
        # get '%' of completeness:
        if len(map_krs) != 0: perc_completeness = len(present)/len(map_krs)*100
        else: perc_completeness = 100   # for maps_noreac
        perc_completeness_str = str(round(perc_completeness))   # version to be printed
        if len(perc_completeness_str)==1: 
            perc_completeness_str = ' ' + perc_completeness_str

            
        # append map to list:
        list_coverage.append({
            'map_id': map_id,
            'map_name_short': map_name_short, 
            'perc_completeness': perc_completeness,
            'perc_completeness_str': perc_completeness_str,
            'present': present,
            'missing': missing,
            'md_ids': [j['md_id'] for j in i['mds']],
        })
        
        
    
    # create coverage dataframe
    if eggnog != '-' and len(eggnog) >= 2:  
        df_coverage = {}
        for i in list_coverage:
            for kr in i['present'].union(i['missing']):
                if kr not in df_coverage.keys(): 
                    df_coverage[kr] = {'map_ids': set()}
                df_coverage[kr]['map_ids'].add(i['map_id'])
        df_coverage = pnd.DataFrame.from_records(df_coverage).T
        df_coverage['modeled'] = False
        for kr, row in df_coverage.iterrows():
            if kr in kr_ids_modeled:
                df_coverage.loc[kr, 'modeled'] = True
        # build strain columns all at once
        df_strains = []  # list of small DataFrames
        for eggfile in eggnog:
            strain = Path(eggfile).stem
            eggset = parse_eggnog(model, eggfile, idcollection_dict)
            col = df_coverage.index.to_series().isin(eggset).astype(int)  # integer: 0 or 1
            df_strains.append(col.rename(strain))
        df_strains = pnd.concat(df_strains, axis=1)
        # sort rows: upper rows are present in more strains
        #df_strains = df_strains.loc[df_strains.sum(axis=1).sort_values(ascending=False).index]   # commented: now in charge of figures.py
        df_coverage = df_coverage.loc[df_strains.index]
        df_coverage = pnd.concat([df_coverage, df_strains], axis=1)
        # split in 2: modeled above, non-modeled below:
        #df_coverage = pnd.concat([df_coverage[df_coverage['modeled']==True], df_coverage[df_coverage['modeled']==False]])   # commented: now in charge of figures.py
    else:  # not interesting in a super-long table without strains in column
        df_coverage = None
    
           
            
    # order list by '%' of completness and print if needed:
    list_coverage = sorted(list_coverage, key=lambda x: x['perc_completeness'], reverse=True)
    for i in list_coverage:
        if progress:
            if focus=='-' or focus in i['md_ids'] or focus==i['map_id']:
                if i['map_id'] in maps_missing or i['map_id'] in maps_partial:
                    logger.info(f"{i['map_id']}: {i['map_name_short']} {i['perc_completeness_str']}% completed, {len(i['present'])} added, {len(i['missing'])} missing.")
        
        
        # get the correspondent pathway element of the 'summary_dict'
        right_item = None
        for k in summary_dict:
            if k['map_id'] == i['map_id']:
                right_item = k
                
                
        # define some counters:
        mds_completed = set()
        mds_noreac = set()
        mds_missing = set()
        mds_partial = set()


        list_coverage_md  = []
        spacer = '    '


        # iterate over each module:
        for z in right_item['mds']:


            # get ID and name: 
            md_id = z['md_id']
            md_name_short = f"{z['md_name'][:20]}"
            if len(z['md_name']) > 20: 
                md_name_short = md_name_short + '...'
            else:  # add spaces as needed: 
                md_name_short = md_name_short + ''.join([' ' for i in range(23-len(md_name_short))])


            # check if this module was (at least partially) covered:
            md_krs = set([kr for kr in z['kr_ids_md'] if kr in kr_uni])
            missing = md_krs - kr_ids_modeled
            present = kr_ids_modeled.intersection(md_krs)
            if focus == md_id: 
                missing_logger = (md_id, missing)
            
            
            # put the map in the right bucket:
            if missing == set() and md_krs != set():
                mds_completed.add(md_id)
            elif md_krs == set():
                mds_noreac.add(md_id)
            elif missing == md_krs:
                mds_missing.add(md_id)
            elif len(missing) < len(md_krs):
                mds_partial.add(md_id)

                
            # get '%' of completeness:
            if len(md_krs) != 0: perc_completeness = len(present)/len(md_krs)*100
            else: perc_completeness = 100   # for mds_noreac
            perc_completeness_str = str(round(perc_completeness))   # version to be printed
            if len(perc_completeness_str)==1: 
                perc_completeness_str = ' ' + perc_completeness_str

                
            # append md to list:
            list_coverage_md.append({
                'md_id': md_id,
                'md_name_short': md_name_short, 
                'perc_completeness': perc_completeness,
                'perc_completeness_str': perc_completeness_str,
                'present': present,
                'missing': missing,
            })
               
            
        # order list by '%' of completness and print if needed:
        list_coverage_md = sorted(list_coverage_md, key=lambda x: x['perc_completeness'], reverse=True)
        for z in list_coverage_md:
            if module:
                if focus=='-' or focus==z['md_id']:
                    if z['md_id'] in mds_missing or z['md_id'] in mds_partial:
                        logger.info(f"{spacer}{z['md_id']}: {z['md_name_short']} {z['perc_completeness_str']}% completed, {len(z['present'])} added, {len(z['missing'])} missing.")
        
        
        # print summary:
        if module and focus=='-':
            logger.info(f"{spacer}Modules of {right_item['map_id']}: completed {len(mds_completed)} - partial {len(mds_partial)} - missing {len(mds_missing)} - noreac {len(mds_noreac)}")
    if focus != '-':
        logger.info(f"Missing reactions focusing on {missing_logger[0]}: {' '.join(list(missing_logger[1]))}.")
    if progress:
        logger.info(f"Maps: finished {len(maps_finished)} - partial {len(maps_partial)} - missing {len(maps_missing)} - noreac {len(maps_noreac)}")
            
        
    return df_coverage     





from .biomass import *
from .downloads import *
from .excelhub import *
from .fluxbal import *
from .medium import *
from .metrics import *
from .sbmlutils import *
from .escherutils import *
from .logutils import *
from .keggutils import *
from .memoteutils import *



import requests
import io
import threading
import time
import tempfile
import json
import glob


import gdown
import pandas as pnd



def get_dbuni(logger):
    
    
    sheet_id = "1dXJBIFjCghrdvQtxEOYlVNWAQU4mK-nqLWyDQeUZqek"
    #sheet_id = "1dCVOOnpNg7rK3iZmTDz3wybW7YrUNoClnqezT9Q5bpc" # alternative
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Retry! If persists, please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['T', 'R', 'M', 'authors']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    dbuni = {}
    dbuni['T'] = exceldb.parse('T')
    dbuni['R'] = exceldb.parse('R')
    dbuni['M'] = exceldb.parse('M')
    dbuni['authors'] = exceldb.parse('authors')
    
    
    # check table headers
    headers = {}
    headers['T'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'author', 'notes']
    headers['R'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'author', 'notes']
    headers['M'] = ['pure_mid', 'formula', 'charge', 'kc', 'name', 'inchikey', 'author', 'notes']
    headers['authors'] = ['username', 'first_name', 'last_name', 'role', 'mail']
    for i in dbuni.keys(): 
        if set(dbuni[i].columns) != set(headers[i]):
            logger.error(f"Sheet '{i}' is missing the columns {set(headers[i]) - set(dbuni[i].columns)}.")
            return 1
        
    return dbuni



def get_dbexp(logger):
    
    
    sheet_id = "1qGbIIipHJgYQjk3M0xDWKvnTkeInPoTeH9unDQkZPwg"
    #sheet_id = "1qxTRf30SeT9WJFYxWm2ChCxkTR0sTn7BbDOFhUuUQIE"   # alternative
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Retry! If persists, please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['media', 'PM1', 'PM2A', 'PM3B', 'PM4A', 'MWF', 'DNA', 'RNA', 'PROTS', 'LIPIDS_PL', 'LIPIDS_FA', 'authors']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    dbexp = {}
    dbexp['media'] = exceldb.parse('media')
    dbexp['PM1'] = exceldb.parse('PM1')
    dbexp['PM2A'] = exceldb.parse('PM2A')
    dbexp['PM3B'] = exceldb.parse('PM3B')
    dbexp['PM4A'] = exceldb.parse('PM4A')
    dbexp['MWF'] = exceldb.parse('MWF')
    dbexp['DNA'] = exceldb.parse('DNA')
    dbexp['RNA'] = exceldb.parse('RNA')
    dbexp['PROTS'] = exceldb.parse('PROTS')
    dbexp['LIPIDS_PL'] = exceldb.parse('LIPIDS_PL')
    dbexp['LIPIDS_FA'] = exceldb.parse('LIPIDS_FA')
    dbexp['authors'] = exceldb.parse('authors')
    
    
    # format tables (media):
    # assign substrates as index
    dbexp['media'].index = dbexp['media'].iloc[:, 1]
    # remove first 2 useless column (empty & substrates)
    dbexp['media'] = dbexp['media'].iloc[:, 2:]
    
    
    # format tables (Biolog(R)):
    for sheet in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        # assign wells as index
        dbexp[sheet].index = dbexp[sheet].iloc[:, 2]
        # remove first 3 useless columns
        dbexp[sheet] = dbexp[sheet].iloc[:, 3:]
        
        
    # format tables (biomass):
    dbexp['MWF'].index = dbexp['MWF'].iloc[:, 0]   # assign index
    dbexp['MWF'] = dbexp['MWF'].iloc[:, 1:]        # remove meaningless columns
    #
    dbexp['DNA'].index = dbexp['DNA'].iloc[:, 0]   # assign index
    dbexp['DNA'] = dbexp['DNA'].iloc[:, 1:]        # remove meaningless columns
    #
    dbexp['RNA'].index = dbexp['RNA'].iloc[:, 0]   # assign index
    dbexp['RNA'] = dbexp['RNA'].iloc[:, 1:]        # remove meaningless columns
    # 
    dbexp['PROTS'].index = dbexp['PROTS'].iloc[:, 1]   # assign index
    dbexp['PROTS'] = dbexp['PROTS'].iloc[:, 2:]        # remove meaningless columns
    #
    dbexp['LIPIDS_PL'].index = dbexp['LIPIDS_PL'].iloc[:, 1]   # assign index
    dbexp['LIPIDS_PL'] = dbexp['LIPIDS_PL'].iloc[:, 2:]        # remove meaningless columns
    #
    dbexp['LIPIDS_FA'].index = dbexp['LIPIDS_FA'].iloc[:, 1]   # assign index
    dbexp['LIPIDS_FA'] = dbexp['LIPIDS_FA'].iloc[:, 2:]        # remove meaningless columns
    
    
    return dbexp



def get_eschermap(logger):
    
    
    lastmap = dict()  
    
    
    folder_id = "1YE4l8IFL9pRgonAmFCf2SMRnpGJCHjil"
    folder_url = f"https://drive.google.com/drive/folders/{folder_id}?usp=sharing"

    
    # the temporary folder (/tmp/on-the-fly-code) will be deleted once exiting the with statement
    with tempfile.TemporaryDirectory() as tmp_dir:
        
        
        # get available versions without downloading: 
        contents = gdown.download_folder(folder_url, output=tmp_dir, quiet=True, skip_download=True)

        if len(contents) == 0:
            logger.error("Online folder 'universe.escher' seems empty. Please contact the developer.")
            return 1
        
        
        # get the last version
        files = {i.path: i.id for i in contents}
        files = dict(sorted(files.items()))   # sort keys in alphabetical order
        last_file_name = list(files.keys())[-1]
        last_file_id = files[last_file_name]
        lastmap['filename'] = last_file_name
        
        
        # download last version:
        try: pathfile_tmpfolder = gdown.download(id=last_file_id, output=f"{tmp_dir}/{last_file_name}", quiet=True)
        except: 
            logger.error("Downloading of last-version universal Escher map failed. Retry. If persists, please contact the developer.")
            return 1
        
        # load json
        with open(pathfile_tmpfolder, 'r') as file:
            json_data = json.load(file)
            lastmap['json'] = json_data

        
    return lastmap
        
    
    
def get_databases(logger): 
    
    
    def run_with_result(func, logger, results_dict, key):
        result = func(logger)
        results_dict[key] = result
        
    
    results_dict = dict()
    t1 = threading.Thread(target=run_with_result, args=(get_dbuni, logger, results_dict, 'dbuni'))
    t2 = threading.Thread(target=run_with_result, args=(get_dbexp, logger, results_dict, 'dbexp'))
    t3 = threading.Thread(target=run_with_result, args=(get_eschermap, logger, results_dict, 'eschermap'))

    
    # wait for the longest download (assess every 0.001 seconds):
    t1.start()
    t2.start()
    t3.start()
    while t1.is_alive() or t2.is_alive() or t3.is_alive():
        time.sleep(0.001)  

        
    if type(results_dict['dbuni'])==int:
        return 1
    if type(results_dict['dbexp'])==int:
        return 1
    if type(results_dict['eschermap'])==int:
        return 1
    return (results_dict['dbuni'], results_dict['dbexp'], results_dict['eschermap'])



def format_expansion(logger, eggnog):
    
    
    # linux/macos usually perform argument axpansion befor passing the parameter to python.
    if type(eggnog) == list: # already expanded by the terminal
        if len(eggnog)==1 and '*' in eggnog[0]:
            original_eggnog = eggnog[0]
            eggnog = glob.glob(eggnog[0])  # glob will append only existing files to the list
            if eggnog == []:
                logger.info(f"No file matching '{original_eggnog}'.")
    
    
    elif type(eggnog) == str:  # in the terminal, it can be specified by using single quotes
        original_eggnog = False
        if eggnog != '-': # user wanted to specify something
            original_eggnog = eggnog
        eggnog = glob.glob(eggnog)  # glob will append only existing files to the list
        if original_eggnog and eggnog == []:
            logger.info(f"No file matching '{original_eggnog}'.")
            
            
    if eggnog == [] or eggnog == ['-']:
        eggnog = '-'   # return always a list except for this case
        
        
    return eggnog
    


def check_taxon(logger, taxon, idcollection_dict):
    
    
    # verify presence of needed assets
    if 'ko_to_taxa' not in idcollection_dict.keys():
        logger.error(f"Asset 'ko_to_taxa' not found in 'gsrap.maps'. Please update 'gsrap.maps' with 'gsrap getmaps'.")
        return 1
    
    
    # extract level and name
    try: level, name = taxon.split(':')
    except: 
        logger.error(f"Provided --taxon is not well formatted: '{taxon}'.")
        return 1
    
    
    # compute available levels and check
    avail_levels = set(['kingdom', 'phylum'])            
    if level not in avail_levels:
        logger.error(f"Provided level is not acceptable: '{level}' (see --taxon). Acceptable levels are {avail_levels}.")
        return 1
    
    
    # compute available taxa at input level
    avail_taxa_at_level = set()
    ko_to_taxa = idcollection_dict['ko_to_taxa']
    for ko in ko_to_taxa.keys():
        for taxon_name in ko_to_taxa[ko][level]:
            avail_taxa_at_level.add(taxon_name)
    if name not in avail_taxa_at_level:
        logger.error(f"Provided taxon name is not acceptable: '{name}' (see --taxon). Acceptable taxon names for level '{level}' are {avail_taxa_at_level}.")
        return 1

    
    return 0
